import static ch.qos.logback.classic.Level.ALL
import static ch.qos.logback.classic.Level.DEBUG
import static ch.qos.logback.classic.Level.INFO
import static ch.qos.logback.classic.Level.WARN
import static ch.qos.logback.classic.Level.ERROR

import java.lang.management.ManagementFactory

import org.slf4j.Marker
import org.slf4j.MarkerFactory

import ch.qos.logback.classic.Level
import ch.qos.logback.classic.Logger
import ch.qos.logback.classic.LoggerContext
import ch.qos.logback.classic.filter.ThresholdFilter 
import ch.qos.logback.classic.encoder.PatternLayoutEncoder
import ch.qos.logback.classic.gaffer.ConfigurationDelegate
import ch.qos.logback.classic.jmx.JMXConfigurator
import ch.qos.logback.classic.jmx.MBeanUtil
import ch.qos.logback.classic.turbo.TurboFilter
import ch.qos.logback.core.ConsoleAppender
import ch.qos.logback.core.rolling.RollingFileAppender
import ch.qos.logback.core.rolling.TimeBasedRollingPolicy
import ch.qos.logback.core.spi.FilterReply
import ch.qos.logback.core.util.Loader
import ch.qos.logback.core.filter.Filter
import ch.qos.logback.classic.spi.ILoggingEvent
import ch.qos.logback.core.spi.FilterReply

import com.clcbio.api.base.logging.ClcLogger

// This file is invoked in the scope of ContextAware, and the following variables and methods are available:
// context -- The current ch.qos.logback.core.Context
// addStatus(Status)
// addInfo(String)
// addInfo(String, Throwable)
// addWarn(String)
// addWarn(String, Throwable)
// addError(String)
// addError(String, Throwable)

// CLASS DEFINITIONS
class GroovyLogConfig {
    static final String DEFAULT_LOG_FORMAT = "%date{'dd/MM/yyyy HH:mm:ss z'} %-5level %c{30}: %message%n"
    static final String DEFAULT_LOG_FORMAT_CALLER_LOGGING = "%date{'dd/MM/yyyy HH:mm:ss z'} %-5level %c{30}, %C.%M\\(%F:%L\\): %message%n";
    static String logFormat = null
    static Closure configuration = null
    static List<Closure> logLevelConfig = null
    static boolean callerLogging = false
    static ConsoleAppender consoleAppender
    static RollingFileAppender fileAppender
    static ThresholdFilter consoleLevelFilter
    static ProgramInfoFilter consoleProgramInfoFilter
}

class ProgramInfoFilter extends Filter<ILoggingEvent> {
    
    private boolean enabled
    
    public ProgramInfoFilter() {
        super()
        enabled = true
    }
    
    FilterReply decide(ILoggingEvent event) {
    	if( !enabled)
    		return FilterReply.NEUTRAL
    		
        if( event.loggerName.equals("com.clcbio.log.Workbench.ProgramInfo") ) {
            return FilterReply.DENY
        }
        return FilterReply.ACCEPT
    }
    
    void enable() {
    	enabled = true;
    }
    
    void disable() {
    	enabled = false;
    }
}

// METHOD DEFINITIONS
def jmxConfigurator() {
    def contextName = context.name
    def objectNameAsString = MBeanUtil.getObjectNameFor(contextName, JMXConfigurator.class)
    def objectName = MBeanUtil.string2ObjectName(context, this, objectNameAsString)
    def platformMBeanServer = ManagementFactory.getPlatformMBeanServer()
    if (!MBeanUtil.isRegistered(platformMBeanServer, objectName)) {
        JMXConfigurator jmxConfigurator = new JMXConfigurator((LoggerContext) context, platformMBeanServer, objectName)
        try {
            platformMBeanServer.registerMBean(jmxConfigurator, objectName)
        } catch (all) {
            addError("Failed to create mbean", all)
        }
    }
}

def includeFile(File file) {
    if( !file.exists() ) {
        return
    }
    println("Including log configuration " + file)
    def emptyScript = new GroovyShell().parse("5")
    def s = new GroovyShell().parse(file.text)
    s.metaClass.mixin(ConfigurationDelegate)
    
    def defaultMethods = emptyScript.getClass().methods
    
    this.class.methods.findAll { !defaultMethods.contains(it) && it.name != "main" && it.name != "run"}.each {
       s.binding[it.name] = new org.codehaus.groovy.runtime.MethodClosure(this, it.name)
    }
    
    s.setContext(context)
    s.run()
}

def includeResource(String resourceName) {
    def res = Loader.getResource(resourceName, Loader.getClassLoaderOfObject(this))
    if( res != null && res.toURI().getScheme() == "file") {
        includeFile(new File(res.toURI()))
    }
}

def resetLoggers() {
    context.reset()
    context.getLogger(Logger.ROOT_LOGGER_NAME).detachAndStopAllAppenders()
}

def setLogFormat(String format) {
    GroovyLogConfig.logFormat = format
}

def setCallerLogging(boolean enable) {
    GroovyLogConfig.callerLogging = enable
}

def configureFromScratch(Closure c) {
    GroovyLogConfig.configuration = c
}

def appendConfiguration(Closure c) {
    if( GroovyLogConfig.logLevelConfig == null ) {
        GroovyLogConfig.logLevelConfig = new ArrayList<Closure>();
    }
    GroovyLogConfig.logLevelConfig.add(c)
}

def getLogger(org.slf4j.Logger log) {
    return (Logger)log;
}

def getLogger(String name) {
    return context.getLogger(name)
}

def setConsoleLogLevel(Level level) {
	GroovyLogConfig.consoleLevelFilter.setLevel(level.toString())
}

def logProgramInfoToConsole(boolean enable) {
	if( enable) {
		GroovyLogConfig.consoleProgramInfoFilter.disable()
	} else {
		GroovyLogConfig.consoleProgramInfoFilter.enable()
	}
}

// MAIN CONFIGURATION

// Enable JMX
jmxConfigurator()

resetLoggers()

def logDirectory = ClcLogger.getDefaultConfiguration().getLogDirectory()
def logName = ClcLogger.getDefaultConfiguration().getLogName()
def fileLoggingEnabled = ClcLogger.getDefaultConfiguration().isFileLoggingEnabled();
def logDirectoryFile = new File(logDirectory)
if( fileLoggingEnabled) {
	logDirectoryFile.mkdirs()
	println("Logging to " + logDirectory)
}

includeResource("clc-log.groovy")
includeFile(new File(System.getProperty("user.home"), "clc-log.groovy"))
includeFile(new File(com.clcbio.api.base.framework.ProgramInfo.getSettingsDir(), "clc-log.groovy"))

if( GroovyLogConfig.configuration != null ) {
    GroovyLogConfig.configuration()
} else {
    if( GroovyLogConfig.logFormat == null) {
        if( GroovyLogConfig.callerLogging) {
            GroovyLogConfig.logFormat = GroovyLogConfig.DEFAULT_LOG_FORMAT_CALLER_LOGGING 
        } else {
            GroovyLogConfig.logFormat = GroovyLogConfig.DEFAULT_LOG_FORMAT
        }
    }
        
	if( fileLoggingEnabled ) {        
    	def logPrefix = new File(logDirectory, logName).getAbsolutePath()     
    
	    appender("File Logger", RollingFileAppender) {
	        encoder(PatternLayoutEncoder) {
	            pattern = GroovyLogConfig.logFormat
	        }
	        file = logPrefix + ".log"
	        append = false
	        rollingPolicy(TimeBasedRollingPolicy) {
	            fileNamePattern = logPrefix + "-%d{yyyy-MM-dd}.log.gz"
	            maxHistory = 5             
	        }
	    }
	    
	    logger(ClcLogger.Root.getName(), ALL, ["File Logger"])
	    GroovyLogConfig.fileAppender = context.getLogger(ClcLogger.Root.getName()).getAppender("File Logger")
	 }
    
    GroovyLogConfig.consoleLevelFilter = new ThresholdFilter()
	GroovyLogConfig.consoleProgramInfoFilter = new ProgramInfoFilter()
               
    appender("Console Logger", ConsoleAppender) {
        encoder(PatternLayoutEncoder) {
            pattern = GroovyLogConfig.logFormat
        }
                
        component.addFilter(GroovyLogConfig.consoleLevelFilter)                
        component.addFilter(GroovyLogConfig.consoleProgramInfoFilter)
    }        
    
    logger(ClcLogger.Root.getName(), INFO, ["Console Logger"])

    GroovyLogConfig.consoleAppender = context.getLogger(ClcLogger.Root.getName()).getAppender("Console Logger")
    context.getLogger(ClcLogger.Root.getName()).setLevel(INFO)

	setConsoleLogLevel(ERROR)

    if( GroovyLogConfig.logLevelConfig != null ) {
        GroovyLogConfig.logLevelConfig.each {
            it.delegate = new GroovyLogConfig()
            it()
        }
    }    
    
    GroovyLogConfig.consoleLevelFilter.start()
}